@@ -402,8 +402,13 @@ impl Einsum {
     #[staticmethod]
     #[args(nodes = "*")]
     pub fn elementwise_broadcasted(nodes: Vec<CircuitRc>, name: Option<String>) -> Result<Einsum> {
-        let out_shape = broadcast_shapes(&nodes.iter().map(|x| x.info().shape.clone()).collect())
-            .context("failed to broadcast for einsum mul")?;
+        let out_shape = broadcast_shapes(
+            &nodes
+                .iter()
+                .map(|x| &x.info().shape[..])
+                .collect::<Vec<_>>(),
+        )
+        .context("failed to broadcast for einsum mul")?;
         let mut prev_one_shape = out_shape.len().saturating_sub(1) as u8;
         Einsum::try_new(
             nodes
@@ -570,7 +575,13 @@ impl Add {
     }
 
     fn compute_shape_maybe(&self) -> Result<Shape> {
-        broadcast_shapes(&self.nodes.iter().map(|x| x.info().shape.clone()).collect())
+        broadcast_shapes(
+            &self
+                .nodes
+                .iter()
+                .map(|x| &x.info().shape[..])
+                .collect::<Vec<_>>(),
+        )
     }
 }
 
@@ -1244,10 +1255,10 @@ impl Index {
                 .0
                 .iter()
                 .map(|x| match x {
-                    TensorAxisIndex::Tensor(t) => t.shape().clone(),
+                    TensorAxisIndex::Tensor(t) => &t.shape()[..],
                     _ => panic!(),
                 })
-                .collect(),
+                .collect::<Vec<_>>(),
         )?;
         let rearranged_back = Rearrange::try_new(
             indexed,
@@ -1768,8 +1779,8 @@ impl Conv {
                 bail!(ConstructError::ConvFilterWrongShape {});
             }?;
         let output_batch_shape =
-            broadcast_shapes(&vec![input_batch_shape.clone(), filter_batch_shape.clone()])
-                .context("input and filter shapes dont broadcast")?;
+            broadcast_shapes(&vec![&input_batch_shape[..], &filter_batch_shape[..]])
+                .context("input and filter batch shapes dont broadcast")?;
 
         if !filter_batch_shape.is_empty() {
             bail!(ConstructError::NotSupportedYet {
