@@ -19,10 +19,9 @@ use uuid::Uuid;
 use crate::{
     circuit_node_auto_impl, circuit_node_extra_impl,
     circuit_node_private::{CircuitNodeComputeInfoImpl, CircuitNodeHashItems},
-    named_axes::set_named_axes,
     new_rc,
     prelude::*,
-    CachedCircuitInfo, NamedAxes, PyCircuitBase, Shape, TensorEvalError,
+    CachedCircuitInfo, PyCircuitBase, Shape, TensorEvalError,
 };
 
 macro_rules! compute_info_auto_leaf_impl {
@@ -143,16 +142,11 @@ impl Array {
 #[pymethods]
 impl Array {
     #[new]
+    #[args(name = "None")]
     fn py_new(value: Tensor, name: Option<String>) -> PyClassInitializer<Self> {
         Array::new(value, name).into_init()
     }
 
-    #[staticmethod]
-    pub fn new_named_axes(value: Tensor, name: Option<String>, named_axes: NamedAxes) -> Self {
-        let result = Self::new(value, name);
-        set_named_axes(result, named_axes)
-    }
-
     #[staticmethod]
     #[pyo3(name = "randn")]
     #[args(
