@@ -1,16 +1,21 @@
-use std::iter::{self, zip};
+use std::{
+    collections::BTreeMap,
+    iter::{self, zip},
+};
 
 use anyhow::{bail, Context, Result};
 use macro_rules_attribute::apply;
 use pyo3::{exceptions::PyValueError, prelude::*};
 use rr_util::{
     pycall, python_error_exception,
-    rearrange_spec::{ExpandToSpecOrShape, OpShape, OpSize},
+    rearrange_spec::{ExpandToSpecOrShape, OpShape, OpSize, RearrangeSpec},
     sv,
     symbolic_size::SymbolicSizeProduct,
-    tensor_util::{Shape, TensorAxisIndex, TensorIndex},
+    tensor_util::{broadcast_shapes, Shape, TensorAxisIndex, TensorIndex},
+    util::EinsumAxes,
 };
 use rustc_hash::FxHashMap as HashMap;
+use smallvec::SmallVec;
 use thiserror::Error;
 
 use crate::{
@@ -21,16 +26,33 @@ use crate::{
 #[pyfunction]
 #[pyo3(name = "expand_node")]
 pub fn expand_node_py(circuit: CircuitRc, inputs: Vec<CircuitRc>) -> Result<CircuitRc> {
-    expand_node(circuit, &inputs)
+    expand_node(circuit, &inputs, |_, _, _| {
+        bail!("fancy module expanding not supported via python expand node")
+    })
 }
 
-pub fn expand_node(circuit: CircuitRc, inputs: &Vec<CircuitRc>) -> Result<CircuitRc> {
+pub fn expand_node(
+    circuit: CircuitRc,
+    inputs: &Vec<CircuitRc>,
+    rerun_with_extra_replacements: impl FnOnce(
+        CircuitRc,
+        HashMap<CircuitRc, CircuitRc>,
+        usize,
+    ) -> Result<CircuitRc>,
+) -> Result<CircuitRc> {
     if inputs.len() != circuit.children().count() {
         bail!(ExpandError::WrongNumChildren {
             expected: circuit.children().count(),
             got: inputs.len(),
         });
     }
+    if inputs
+        .iter()
+        .zip(circuit.children())
+        .all(|(new, old)| new == &old)
+    {
+        return Ok(circuit); // in all identical case we can return same circuit for efficiency
+    }
     let batch_ranks: Vec<usize> = zip(circuit.children(), inputs)
         .filter_map(|(old, new)| new.info().rank().checked_sub(old.info().rank()))
         .collect();
@@ -41,12 +63,46 @@ pub fn expand_node(circuit: CircuitRc, inputs: &Vec<CircuitRc>) -> Result<Circui
         });
     }
 
-    // TODO: maybe we should allow for inconsistent symbolic batch shapes?
-    // (probably not...)
     let batch_shapes: Vec<&[usize]> = zip(&batch_ranks, inputs)
         .map(|(br, new)| &new.info().shape[0..*br])
         .collect();
 
+    // TODO: maybe we should allow for inconsistent symbolic batch shapes?
+    // (probably not...)
+    let broadcasted_batch_shape =
+        broadcast_shapes(&batch_shapes).context("batch shapes couldn't be broadcast together")?;
+    let batch_rank = broadcasted_batch_shape.len();
+
+    let get_expanded_inputs = |batchable: &[bool]| -> Result<Vec<CircuitRc>, usize> {
+        let br = batch_rank;
+        assert_eq!(batchable.len(), batch_ranks.len());
+        assert_eq!(inputs.len(), batch_ranks.len());
+        inputs
+            .iter()
+            .zip(&batch_ranks)
+            .zip(batchable)
+            .enumerate()
+            .map(|(i, ((x, &r), batchable_x))| {
+                assert!(br >= r);
+                if !batchable_x && r > 0 {
+                    return Err(i);
+                }
+                if br == r || !batchable_x {
+                    return Ok(x.clone());
+                }
+                Ok(Rearrange::nrc(
+                    x.clone(),
+                    RearrangeSpec::prepend_batch_shape(
+                        SmallVec::from_slice(&broadcasted_batch_shape[..br - r]),
+                        x.ndim(),
+                    )
+                    .unwrap(),
+                    x.name().map(|x| format!("{} rep_for_batch", x)),
+                ))
+            })
+            .collect()
+    };
+
     let find_non_symbolic = |sizes: &[usize]| {
         sizes
             .into_iter()
@@ -54,7 +110,7 @@ pub fn expand_node(circuit: CircuitRc, inputs: &Vec<CircuitRc>) -> Result<Circui
             .unwrap_or(sizes[0])
     };
 
-    match &**circuit {
+    let out = match &**circuit {
         Circuit::SetSymbolicShape(_) => Ok(inputs[0].clone()), /* just return circuit, symbolic shapes will be added as needed! (TODO: maybe should keep?) */
         Circuit::Symbol(_) | Circuit::Scalar(_) => Ok(circuit.clone()),
         Circuit::Rearrange(node) => {
@@ -151,19 +207,13 @@ pub fn expand_node(circuit: CircuitRc, inputs: &Vec<CircuitRc>) -> Result<Circui
             ))
         }
         Circuit::Concat(node) => {
-            // TODO: set symbolic size, error on concat axis having symbolic size? (or maybe concat axis is fine???
-            if !batch_shapes.iter().all(|x| x == &batch_shapes[0]) {
-                bail!(ExpandError::InconsistentBatches {
-                    batch_shapes: batch_shapes
-                        .iter()
-                        .map(|x| x.iter().cloned().collect())
-                        .collect(),
-                    circuit: circuit.clone(),
-                });
-            }
-            let br = batch_ranks[0];
+            // TODO: error on concat axis having symbolic size? (or maybe concat axis is fine???
+            let br = batch_rank;
+
+            let inputs = get_expanded_inputs(&vec![true; inputs.len()]).unwrap();
+
             let new_axis = node.axis + br;
-            if !zip(&node.nodes, inputs)
+            if !zip(&node.nodes, &inputs)
                 .all(|(old, new)| old.info().shape[node.axis] == new.info().shape[new_axis])
             {
                 bail!(ExpandError::ConcatAxis {
@@ -252,48 +302,48 @@ pub fn expand_node(circuit: CircuitRc, inputs: &Vec<CircuitRc>) -> Result<Circui
             Add::try_new(inputs, node.name_cloned()).map(|x| x.rc())
         }
         Circuit::GeneralFunction(node) => {
-            // TODO: symbolic!
-            GeneralFunction::try_new(inputs.clone(), node.spec.clone(), node.name_cloned())
-                .map(|x| x.rc())
+            // TODO: symbolic? (annoying to handle, would require changes to spec)
+            // We could optimize out these input expansions in various cases.
+            // TODO: we could run into strange issues with general functions and expand.
+            // Maybe resolve with an assert...
+            // TODO: maybe alow for batching over individual inputs like:
+            // (This is a bit annoying in various ways and probably shouldn't be supported.
+            // - [b_0, b_1, b_2, (stuff)]
+            // - [b_2, (other stuff)]
+            GeneralFunction::try_new(
+                get_expanded_inputs(&node.input_batchability).map_err(|input_i| {
+                    ExpandError::GeneralFunctionTriedToBatchNonBatchableInput {
+                        input_i,
+                        batched_inputs: inputs.clone(),
+                        general_function: node.clone(),
+                    }
+                })?,
+                node.spec.clone(),
+                node.name_cloned(),
+            )
+            .map(|x| x.rc())
         }
         Circuit::Einsum(node) => {
-            let mut batch_shape: Option<&[usize]> = None;
-            for bs in &batch_shapes {
-                if !bs.is_empty() {
-                    if let Some(existing) = batch_shape {
-                        if *bs != existing {
-                            bail!(ExpandError::InconsistentBatches {
-                                batch_shapes: batch_shapes
-                                    .iter()
-                                    .map(|x| x.iter().cloned().collect())
-                                    .collect(),
-                                circuit: circuit.clone(),
-                            });
-                        }
-                    } else {
-                        batch_shape = Some(bs.clone());
-                    }
-                }
-            }
+            let br = batch_rank;
             let next_axis = node.next_axis();
-            let newies = || (next_axis as u8..next_axis + batch_shape.unwrap().len() as u8);
-            let out_axes = if let Some(_bs) = batch_shape {
-                newies().chain(node.out_axes.iter().cloned()).collect()
-            } else {
-                node.out_axes.clone()
-            };
+            assert!(next_axis as usize + br <= u8::MAX as usize);
+            let end_axis = next_axis + br as u8;
+            let out_axes = (next_axis..end_axis)
+                .chain(node.out_axes.iter().cloned())
+                .collect();
             let new_args: Vec<_> = node
                 .args
                 .iter()
-                .enumerate()
-                .map(|(i, (_child, ints))| {
-                    (inputs[i].clone(), {
-                        if !batch_shapes[i].is_empty() {
-                            newies().chain(ints.iter().cloned()).collect()
-                        } else {
-                            ints.clone()
-                        }
-                    })
+                .zip(inputs)
+                .zip(batch_ranks)
+                .map(|(((_child, ints), inp), r)| {
+                    assert!(br >= r);
+                    (
+                        inp.clone(),
+                        (next_axis + (br - r) as u8..end_axis)
+                            .chain(ints.iter().cloned())
+                            .collect::<EinsumAxes>(),
+                    )
                 })
                 .collect();
 
@@ -347,40 +397,227 @@ pub fn expand_node(circuit: CircuitRc, inputs: &Vec<CircuitRc>) -> Result<Circui
             let rest = children.split_off(1);
             let spec_circuit = children.pop().unwrap();
 
-            // TODO: maybe dedup with parse and other places!
-            let (arg_specs, nodes) = rest
+            if node.nodes.is_empty() {
+                // handle empty case so we can assume non-empty below
+                return Ok(Module::nrc(
+                    vec![],
+                    ModuleSpec {
+                        circuit: spec_circuit,
+                        arg_specs: vec![],
+                    },
+                    node.name_cloned(),
+                ));
+            }
+
+            let new_nodes = rest
                 .chunks_exact(2)
                 .zip(&node.spec.arg_specs)
                 .map(|(sym_inp, arg_spec)| {
                     if let [sym, inp] = sym_inp {
-                        let symbol = if let Some(symbol) = sym.as_symbol() {
-                            symbol.clone()
-                        } else {
-                            bail!(ExpandError::ModuleSpecArgSymbolReplacedWithNonSymbol {
+                        if sym != &arg_spec.symbol.crc() {
+                            bail!(ExpandError::ModuleArgSpecSymbolChangedInExpand {
                                 old_symbol: arg_spec.symbol.clone(),
-                                new_circuit: sym.clone(),
+                                new_symbol_circ: sym.clone(),
+                                module: node.clone(),
                             });
-                        };
-                        Ok((
-                            ModuleArgSpec {
-                                symbol,
-                                ..arg_spec.clone()
-                            },
-                            inp.clone(),
-                        ))
+                        }
+                        Ok(inp.clone())
                     } else {
                         unreachable!()
                     }
                 })
-                .collect::<Result<Vec<_>>>()?
-                .into_iter()
-                .unzip();
+                .collect::<Result<Vec<_>>>()?;
+
+            // Approach:
+            // suppose that orig node shapes are:
+            // [     b_1, b_2, | x_0, x_1]
+            // [          b_2, | y_0]
+            // [b_0, b_1, b_2, | z_0, z_1]
+            // Where x/y/z is the part which matches up with the input symbol
+            // (with a | to indicate the division) and the `b_i` are batched
+            // over by the module itself. (I aligned batch shapes in the
+            // diagram for clarity, the number of batch dims is different for
+            // each input)
+            //
+            // Suppose this yields an overall module shape of
+            // [b_0, b_1, b_2, | w_0, w_1, w_2, w_3]
+            // Where w_0, ..., w_3 are from the spec circuit itself and b_0, b_1, b_2
+            // are the batching done by the module.
+            //
+            //
+            // Now suppose we get new shapes:
+            // [               n_3,      b_1, b_2, x_0, x_1]
+            // [          n_2, n_3,           b_2, y_0]
+            // [n_0, n_1, n_2, n_3, b_0, b_1, b_2, z_0, z_1]
+            // (n is for 'new')
+            //
+            // Then we'll first move new dims into the symbol part as needed to
+            // make batching 'contiguous' (via a rearrange)
+            // [               b_1, b_2, | n_3, x_0, x_1]
+            // [                    b_2, | n_2, n_3, y_0]
+            // [n_0, n_1, b_0, b_1, b_2, | n_2, n_3, z_0, z_1]
+            //
+            // Note this requires changing the symbol shapes - we'll have to prepend the extra new dims.
+            //
+            // Now the module will originally have the following shape because we shifted dims around:
+            // [n_0, n_1, b_0, b_1, b_2, | n_2, n_3, w_0, w_1, w_2, w_3]
+            //
+            // So we'll 'unshift' these dims to yield:
+            // [n_0, n_1, n_2, n_3, b_0, b_1, b_2,  w_0, w_1, w_2, w_3]
+
+            let current_node_batch_ranks: Vec<_> = node
+                .nodes
+                .iter()
+                .zip(&node.spec.arg_specs)
+                .map(|(node, arg_spec)| node.ndim().checked_sub(arg_spec.symbol.ndim()).unwrap())
+                .collect();
+            let current_batch_rank = *current_node_batch_ranks.iter().max().unwrap();
+
+            let new_node_batch_ranks: Vec<_> = new_nodes
+                .iter()
+                .zip(&node.nodes)
+                .map(|(node, orig_node)| node.ndim().checked_sub(orig_node.ndim()).unwrap())
+                .collect();
+
+            let spec_circuit_batch_rank = batch_ranks[0]; // spec circuit is 0th child
+
+            // how many dims do we have to move to make batching 'contiguous'
+            let rank_to_push_into_spec_circuit = current_node_batch_ranks
+                .iter()
+                .zip(&new_node_batch_ranks)
+                .map(|(this_current_batch_rank, this_new_batch_rank)| {
+                    if *this_current_batch_rank == current_batch_rank {
+                        0
+                    } else {
+                        *this_new_batch_rank
+                    }
+                })
+                .chain(iter::once(spec_circuit_batch_rank))
+                .max()
+                .unwrap();
+
+            let (spec_circuit, arg_specs, new_nodes) = if rank_to_push_into_spec_circuit > 0 {
+                let to_unzip = node
+                    .spec
+                    .arg_specs
+                    .iter()
+                    .zip(new_node_batch_ranks)
+                    .zip(new_nodes)
+                    .zip(current_node_batch_ranks)
+                    .map(
+                        |(((arg_spec, this_new_batch_rank), node), this_current_batch_rank)| {
+                            let this_sym_extra_rank =
+                                rank_to_push_into_spec_circuit.min(this_new_batch_rank);
+                            // move dims to make batching contiguous
+                            let new_sym = Symbol::new(
+                                node.shape()[this_new_batch_rank - this_sym_extra_rank
+                                    ..this_new_batch_rank]
+                                    .iter()
+                                    .chain(arg_spec.symbol.shape())
+                                    .cloned()
+                                    .collect(),
+                                arg_spec.symbol.uuid,
+                                arg_spec.symbol.name_cloned(),
+                            );
+                            let node = if this_current_batch_rank > 0 && this_sym_extra_rank > 0 {
+                                let old_sym_rank = arg_spec.symbol.rank();
+                                let node_rank = node.rank();
+                                let up_to_old_sym_rank = node_rank - old_sym_rank;
+                                let shift_name = node.name().map(|x| format!("{} shift_batch", x));
+                                let spec = RearrangeSpec::new_permute(
+                                    (this_sym_extra_rank..up_to_old_sym_rank)
+                                        .chain(0..this_sym_extra_rank)
+                                        .chain(up_to_old_sym_rank..node_rank)
+                                        .collect(),
+                                )
+                                .unwrap();
+                                assert!(!spec.is_identity()); // if statement should be checking for identity
+                                Rearrange::nrc(node, spec, shift_name)
+                            } else {
+                                node
+                            };
+
+                            (
+                                (arg_spec.symbol.crc(), new_sym.crc()),
+                                ModuleArgSpec {
+                                    symbol: new_sym,
+                                    ..arg_spec.clone()
+                                },
+                                node,
+                            )
+                        },
+                    );
+                let (replacements, new_arg_specs, new_nodes): (Vec<_>, _, _) =
+                    itertools::multiunzip(to_unzip);
+                let replacements = replacements // this is just for efficiency
+                    .into_iter()
+                    .filter_map(|(a, b)| (a != b).then_some((a, b)))
+                    .collect();
+
+                let new_spec_circuit =
+                    rerun_with_extra_replacements(node.spec.circuit.clone(), replacements, 0)
+                        .with_context(|| {
+                            format!("failed to rerun batching for spec circuit with expanded symbols! (TODO: info)")
+                        })?;
+
+                assert_eq!(
+                    new_spec_circuit.ndim(),
+                    rank_to_push_into_spec_circuit + node.spec.circuit.ndim(),
+                    "this assumption should hold due to guarantees that batching makes!"
+                );
+
+                (new_spec_circuit, new_arg_specs, new_nodes)
+            } else {
+                (spec_circuit, node.spec.arg_specs.clone(), new_nodes)
+            };
 
             let spec = ModuleSpec {
                 circuit: spec_circuit,
                 arg_specs,
             };
-            Module::try_new(nodes, spec, node.name_cloned()).map(|z| z.rc())
+
+            let out = Module::try_new(new_nodes, spec, node.name_cloned())?.rc();
+
+            // unshift dims
+            let final_out = if current_batch_rank > 0 && rank_to_push_into_spec_circuit > 0 {
+                let orig_spec_ndim = node.spec.circuit.ndim();
+                assert_eq!(
+                    out.ndim(),
+                    current_batch_rank + batch_rank + orig_spec_ndim,
+                    "this assumption should hold due to guarantees that batching makes!"
+                );
+                assert!(batch_rank >= rank_to_push_into_spec_circuit);
+                let batch_rank_done_via_mod = batch_rank - rank_to_push_into_spec_circuit;
+                let total_new_old_batch_via_mod = batch_rank_done_via_mod + current_batch_rank;
+
+                // starting to get a bit complicated : )
+                let new_batch_dims_via_mod = 0..batch_rank_done_via_mod;
+                let end_all_batching = total_new_old_batch_via_mod + rank_to_push_into_spec_circuit;
+                let new_batch_dims_via_push_into_spec_circuit =
+                    total_new_old_batch_via_mod..end_all_batching;
+                let old_batch_dims = batch_rank_done_via_mod..total_new_old_batch_via_mod;
+                let remaining_spec_circuit_dims =
+                    end_all_batching..end_all_batching + orig_spec_ndim;
+                assert_eq!(end_all_batching + orig_spec_ndim, out.ndim());
+
+                let spec = RearrangeSpec::new_permute(
+                    new_batch_dims_via_mod
+                        .chain(new_batch_dims_via_push_into_spec_circuit)
+                        .chain(old_batch_dims)
+                        .chain(remaining_spec_circuit_dims)
+                        .collect(),
+                )
+                .unwrap();
+                assert!(!spec.is_identity()); // if statement should be checking for identity
+
+                let shift_name = out.name().map(|x| format!("{} shift_batch", x));
+                let out_name = out.name_cloned();
+                Rearrange::nrc(out.rename(shift_name), spec, out_name)
+            } else {
+                out
+            };
+
+            Ok(final_out)
         }
         Circuit::Tag(node) => Ok(Tag::new(inputs[0].clone(), node.uuid, node.name_cloned()).rc()),
         _ => {
@@ -392,7 +629,12 @@ pub fn expand_node(circuit: CircuitRc, inputs: &Vec<CircuitRc>) -> Result<Circui
                 })
             }
         }
-    }
+    }?;
+    assert!(
+        out.ndim() == circuit.ndim() + batch_rank,
+        "batching assumption violated!"
+    );
+    Ok(out)
 }
 
 #[apply(python_error_exception)]
@@ -425,35 +667,22 @@ pub enum ExpandError {
         new_shape: Shape,
     },
 
-    #[error("Inputs that should have same batching have different batchings, {batch_shapes:?} {circuit:?} ({e_name})")]
-    InconsistentBatches {
-        batch_shapes: Vec<Shape>,
-        circuit: CircuitRc,
+    // error could be improved...
+    #[error("input_i={input_i} is not batchable, but tried to batch\nbatched_inputs={batched_inputs:?} general_function={general_function:?}\n ({e_name})")]
+    GeneralFunctionTriedToBatchNonBatchableInput {
+        input_i: usize,
+        batched_inputs: Vec<CircuitRc>,
+        general_function: GeneralFunction,
     },
 
     #[error("trying to expand node, unknown variant {variant} ({e_name})")]
     NodeUnhandledVariant { variant: String },
 
-    #[error("Not currently supported! old_shape={old_shape:?} != new_shape={new_shape:?}\nold_spec_circuit={old_spec_circuit:?} new_spec_circuit={new_spec_circuit:?} ({e_name})")]
-    ModuleSpecCircuitChangedShapeInExpand {
-        old_shape: Shape,
-        new_shape: Shape,
-        old_spec_circuit: CircuitRc,
-        new_spec_circuit: CircuitRc,
-    },
-
-    #[error("old_symbol={old_symbol:?} new_circuit={new_circuit:?} ({e_name})")]
-    ModuleSpecArgSymbolReplacedWithNonSymbol {
+    #[error("Not currently supported! old_symbol={old_symbol:?} != new_symbol_circ={new_symbol_circ:?}\nmodule={module:?} ({e_name})")]
+    ModuleArgSpecSymbolChangedInExpand {
         old_symbol: Symbol,
-        new_circuit: CircuitRc,
-    },
-
-    #[error("Not currently supported! old_shape={old_shape:?} != new_shape={new_shape:?}\nold_symbol={old_symbol:?} new_symbol={new_symbol:?} ({e_name})")]
-    ModuleSpecArgSymbolChangedShapeInExpand {
-        old_shape: Shape,
-        new_shape: Shape,
-        old_symbol: Symbol,
-        new_symbol: Symbol,
+        new_symbol_circ: CircuitRc,
+        module: Module,
     },
 
     #[error("node_rank={node_rank} < symbol_rank={symbol_rank}, arg_spec={arg_spec:?} node_shape={node_shape:?} spec_circuit={spec_circuit:?} ({e_name})")]
@@ -509,6 +738,17 @@ pub fn replace_expand_bottom_up_py(circuit: CircuitRc, f: PyObject) -> Result<Ci
 }
 
 pub fn replace_expand_bottom_up<F>(circuit: CircuitRc, replacer: F) -> Result<CircuitRc>
+where
+    F: Fn(CircuitRc) -> Option<CircuitRc>,
+{
+    replace_expand_bottom_up_impl(circuit, &replacer, Default::default())
+}
+
+pub fn replace_expand_bottom_up_impl<F>(
+    circuit: CircuitRc,
+    replacer: &F,
+    extra_replacements: BTreeMap<CircuitRc, CircuitRc>,
+) -> Result<CircuitRc>
 where
     F: Fn(CircuitRc) -> Option<CircuitRc>,
 {
@@ -516,7 +756,37 @@ where
         if let Some(replaced) = replacer(circuit.clone()) {
             return Ok(replaced);
         }
-        expand_node(circuit, new_children)
+        if let Some(replaced) = extra_replacements.get(&circuit) {
+            return Ok(replaced.clone());
+        }
+        expand_node(circuit, new_children, |c, rep, _child_idx| {
+            // This recursion is a bit gross + uncached (caching requires caching by extra_replacements ofc, so a bit more complex) : /
+            replace_expand_bottom_up_impl(
+                c,
+                replacer,
+                extra_replacements
+                    .iter()
+                    .map(|(a, b)| (a.clone(), b.clone()))
+                    .chain(rep)
+                    .collect(),
+            )
+        })
     };
     deep_map_fallible_pre_new_children(circuit, recursor)
 }
+
+#[test]
+fn check_map_override() {
+    let m: BTreeMap<_, _> = [(1, 2), (3, 4), (5, 7), (3, 9)].into_iter().collect();
+    assert_eq!(m[&3], 9);
+    let m: BTreeMap<_, _> = [(1, 2), (3, 4), (5, 7), (1, 9)].into_iter().collect();
+    assert_eq!(m[&1], 9);
+    let m: BTreeMap<_, _> = [(1, 2), (3, 4), (5, 7), (1, 9)].into_iter().collect();
+    assert_eq!(m[&5], 7);
+    let m: BTreeMap<_, _> = [(1, 2), (3, 4), (5, 7), (1, 9), (5, 4), (5, 8), (3, 7)]
+        .into_iter()
+        .collect();
+    assert_eq!(m[&1], 9);
+    assert_eq!(m[&5], 8);
+    assert_eq!(m[&3], 7);
+}
