@@ -1,7 +1,7 @@
 #![feature(let_chains)]
 #![allow(clippy::borrow_deref_ref)]
 use std::{
-    collections::{BTreeMap, BTreeSet},
+    collections::BTreeSet,
     fmt::{self, Debug},
     hash::Hash,
     iter::zip,
@@ -20,6 +20,7 @@ pub use generalfunction::{GeneralFunction, GeneralFunctionSpec};
 use itertools::Itertools;
 use macro_rules_attribute::apply;
 pub use module::{Module, ModuleArgSpec, ModuleSpec};
+use named_axes::set_named_axes;
 use num_bigint::BigUint;
 use opaque_iterative_matcher::OpaqueIterativeMatcherVal;
 use print::PrintOptions;
@@ -40,7 +41,7 @@ use rr_util::{
     symbolic_size::{SymbolicSizeConstraints, SymbolicSizeProduct},
     tensor_util::{check_canon_idxs, Shape, TensorIndex, TorchDeviceDtype, TorchDeviceDtypeOp},
     util::{
-        arc_ref_clone, arc_unwrap_or_clone, AsOp, AxisInt, EinsumAxes, HashBytes, HashBytesToPy,
+        arc_ref_clone, arc_unwrap_or_clone, AsOp, EinsumAxes, HashBytes, HashBytesToPy, NamedAxes,
     },
 };
 use rustc_hash::{FxHashMap as HashMap, FxHashSet as HashSet};
@@ -81,8 +82,6 @@ fn check_canon_axes_as(rank: usize, axes: &[i64]) -> Result<RInnerInts> {
         .collect())
 }
 
-pub type NamedAxes = BTreeMap<AxisInt, String>;
-
 pub trait CircuitNodeSelfOnlyHash {
     fn compute_self_only_hash(&self, hasher: &mut blake3::Hasher);
 }
@@ -2050,6 +2049,11 @@ impl PyCircuitBase {
     fn rename(&self, name: Option<String>) -> CircuitRc {
         self.0.clone().rename(name)
     }
+
+    fn rename_axes(&self, named_axes: NamedAxes) -> Result<CircuitRc> {
+        set_named_axes(self.0.clone(), named_axes)
+    }
+
     fn add_suffix(&self, suffix: Option<&str>) -> CircuitRc {
         self.0.clone().add_suffix(suffix)
     }
