@@ -181,9 +181,12 @@ impl ModuleSpec {
         .with_context(|| {
             format!(
                 concat!(
-                    "replacing symbols with nodes (which are not necessarily symbols)",
+                    "replacing symbols with nodes",
                     " failed in ModuleSpec substitute\n",
-                    "(perhaps you replaced the input symbols to a nested module?)\n",
+                    "common errors:\n",
+                    "  - Multiple dtypes (see below error)\n",
+                    "  - Replacing inputs symbols to a nest module with ",
+                    "non-symbols (nodes aren't necessary symbols)\n",
                     "module_spec={:?}\nnodes={:?}"
                 ),
                 self, nodes
@@ -222,13 +225,13 @@ impl ModuleSpec {
 
 #[pyfunction]
 pub fn all_free(circuit: CircuitRc) -> Vec<Symbol> {
+    // returns syms in circuit visit order (order maybe matters here)
     let mut all_syms = Vec::new();
     visit_circuit_unwrap(substitute_all_modules(circuit), |x| {
         if let Some(sym) = x.as_symbol() {
             all_syms.push(sym.clone())
         }
     });
-    all_syms.sort();
     all_syms
 }
 
