@@ -1,23 +1,39 @@
 use std::collections::BTreeMap;
 
-use anyhow::Result;
-use rr_util::util::is_unique;
+use anyhow::{bail, Result};
+use macro_rules_attribute::apply;
+use pyo3::exceptions::PyRuntimeError;
+use rr_util::{
+    python_error_exception,
+    util::{is_unique, NamedAxes},
+};
 use rustc_hash::{FxHashMap as HashMap, FxHashSet as HashSet};
+use thiserror::Error;
 use uuid::Uuid;
 
 use crate::{
     circuit_utils::deep_map_pass_down_branching, deep_map_unwrap, prelude::*,
-    visit_circuit_with_parents_fallible, NamedAxes,
+    visit_circuit_with_parents_fallible,
 };
 
-pub fn set_named_axes<T: CircuitNode>(node: T, named_axes: NamedAxes) -> T {
-    node.update_info(|i| i.named_axes = named_axes).unwrap()
+pub fn set_named_axes<T: CircuitNode>(node: T, named_axes: NamedAxes) -> Result<T> {
+    for name in named_axes.values() {
+        if name
+            .chars()
+            .any(|c| c.is_whitespace() || c == ']' || c == '[' || c == ':')
+        {
+            bail!(NamedAxesError::ForbiddenCharacter {
+                name: name.to_owned()
+            })
+        }
+    }
+    Ok(node.update_info(|i| i.named_axes = named_axes).unwrap())
 }
 use pyo3::prelude::*;
 
 #[pyfunction]
 #[pyo3(name = "set_named_axes")]
-pub fn set_named_axes_py(circuit: CircuitRc, named_axes: NamedAxes) -> CircuitRc {
+pub fn set_named_axes_py(circuit: CircuitRc, named_axes: NamedAxes) -> Result<CircuitRc> {
     set_named_axes(circuit, named_axes)
 }
 pub fn merge_named_axes(a: &NamedAxes, b: &NamedAxes) -> NamedAxes {
@@ -77,6 +93,7 @@ pub fn propagate_named_axes(
                 circuit.map_children_unwrap_idxs(|i| children[i].clone()),
                 merge_named_axes(&circuit.info().named_axes, named_axes),
             )
+            .unwrap() // Ok because inputs don't have spaces
             .rc()
         },
         named_axes,
@@ -167,6 +184,15 @@ pub fn deep_strip_axis_names(circuit: CircuitRc, names: &Option<HashSet<String>>
         } else {
             BTreeMap::new()
         };
-        set_named_axes(x, axis_names)
+        set_named_axes(x, axis_names).unwrap() // Ok because we are removing names
     })
 }
+
+#[apply(python_error_exception)]
+#[base_error_name(NamedAxes)]
+#[base_exception(PyRuntimeError)]
+#[derive(Error, Debug, Clone)]
+pub enum NamedAxesError {
+    #[error("No whitespace, colon, or square brackets allowed in axis name '{name}'\n({e_name})")]
+    ForbiddenCharacter { name: String },
+}
