@@ -1,6 +1,7 @@
 use std::{iter::zip, str::FromStr};
 
 use anyhow::{bail, Context, Result};
+use itertools::Itertools;
 use macro_rules_attribute::apply;
 use once_cell::sync::Lazy;
 use pyo3::{exceptions::PyValueError, prelude::*};
@@ -12,20 +13,20 @@ use rr_util::{
     tensor_util::{
         parse_numeric, ParseError, PyParseError, Shape, TensorIndex, TorchDeviceDtypeOp,
     },
-    util::{counts_g_1, flip_op_result, is_unique},
+    util::{counts_g_1, flip_op_result, is_unique, NamedAxes},
 };
 use rustc_hash::FxHashMap as HashMap;
-use smallvec::SmallVec as Sv;
 use thiserror::Error;
 use uuid::Uuid;
 
 use crate::{
-    print::TerseBool, Add, Array, CircuitNode, CircuitRc, CircuitType, Concat, ConstructError,
-    Cumulant, DiscreteVar, Einsum, GeneralFunction, Index, Module, ModuleArgSpec, ModuleSpec,
-    Rearrange, Scalar, Scatter, SetSymbolicShape, StoredCumulantVar, Symbol, Tag,
+    named_axes::set_named_axes, print::TerseBool, Add, Array, CircuitNode, CircuitRc, CircuitType,
+    Concat, ConstructError, Cumulant, DiscreteVar, Einsum, GeneralFunction, Index, Module,
+    ModuleArgSpec, ModuleSpec, Rearrange, Scalar, Scatter, SetSymbolicShape, StoredCumulantVar,
+    Symbol, Tag,
 };
 
-const FANCY_PREFIX: &str = "fancy:"; // Should only contain ascii characters
+const FANCY_PREFIX: &str = "fancy:";
 
 #[pyclass]
 #[derive(Clone, Debug)]
@@ -60,10 +61,44 @@ impl Default for Parser {
     }
 }
 
+pub(super) fn get_reference_circuits(
+    reference_circuits: HashMap<String, CircuitRc>,
+    reference_circuits_by_name: Vec<CircuitRc>,
+) -> Result<HashMap<String, CircuitRc>> {
+    let ref_circ_names = reference_circuits_by_name
+        .into_iter()
+        .map(|circ| {
+            Ok((
+                circ.name_cloned()
+                    .ok_or_else(|| ReferenceCircError::ByNameHasNoneName {
+                        circuit: circ.clone(),
+                    })?,
+                circ,
+            ))
+        })
+        .collect::<Result<Vec<_>>>()?;
+
+    let all_idents: Vec<_> = ref_circ_names
+        .iter()
+        .map(|(name, _)| name)
+        .chain(reference_circuits.keys().map(|x| x))
+        .collect();
+    if !is_unique(&all_idents) {
+        bail!(ReferenceCircError::DuplicateIdentifier {
+            dup_idents: counts_g_1(all_idents.into_iter().cloned())
+        })
+    }
+
+    Ok(ref_circ_names
+        .into_iter()
+        .chain(reference_circuits)
+        .collect())
+}
+
 #[pymethods]
 impl Parser {
     #[args(
-        reference_circuits = "HashMap::default()",
+        reference_circuits = "Parser::default().reference_circuits",
         reference_circuits_by_name = "vec![]",
         tensors_as_random = "Parser::default().tensors_as_random",
         tensors_as_random_device_dtype = "Parser::default().tensors_as_random_device_dtype",
@@ -83,38 +118,11 @@ impl Parser {
         module_check_all_args_present: bool,
         module_check_unique_arg_names: bool,
     ) -> Result<Self> {
-        let ref_circ_names = reference_circuits_by_name
-            .into_iter()
-            .map(|circ| {
-                Ok((
-                    circ.name_cloned().ok_or_else(|| {
-                        ParseArgError::ReferenceCircuitByNameHasNoneName {
-                            circuit: circ.clone(),
-                        }
-                    })?,
-                    circ,
-                ))
-            })
-            .collect::<Result<Vec<_>>>()?;
-
-        let all_idents: Vec<_> = ref_circ_names
-            .iter()
-            .map(|(name, _)| name)
-            .chain(reference_circuits.keys().map(|x| x))
-            .collect();
-        if !is_unique(&all_idents) {
-            bail!(ParseArgError::ReferenceCircuitDuplicateIdentifier {
-                dup_idents: counts_g_1(all_idents.into_iter().cloned())
-            })
-        }
-
-        let reference_circuits: HashMap<_, _> = ref_circ_names
-            .into_iter()
-            .chain(reference_circuits)
-            .collect();
-
         Ok(Self {
-            reference_circuits,
+            reference_circuits: get_reference_circuits(
+                reference_circuits,
+                reference_circuits_by_name,
+            )?,
             tensors_as_random,
             tensors_as_random_device_dtype,
             allow_hash_with_random,
@@ -185,6 +193,7 @@ struct PartialCirc {
     pub shape: Option<Shape>,
     pub name: Option<String>,
     pub children: Vec<CircuitIdent>,
+    pub named_axes: Option<NamedAxes>,
 }
 
 impl Parser {
@@ -217,6 +226,7 @@ impl Parser {
         let mut top_level = vec![];
         const WHITESPACE: &str = r"[\s│└├‣─]*";
         const NAME_INSIDE_MATCH: &str = r"(?:(?:\\')?[^']*)*";
+        const AXIS_NAME_MATCH: &str = r"[^\s:\[\]]*"; // No whitespace, colon, or square brackets
 
         let (
             prefix_whitespace_cap,
@@ -229,10 +239,9 @@ impl Parser {
         ) = (1, 2, 3, 4, 5, 6, 7);
         static RE_STR: Lazy<(Regex, String)> = Lazy::new(|| {
             let name_match_capture: String = format!("'({})'", NAME_INSIDE_MATCH);
-            let name_match: String = format!("'{}'", NAME_INSIDE_MATCH);
             // first match named axis name, then size. Name is optional. We handle whitespace.
             let shape_axis_match: String =
-                format!(r"(?:{}\s*:\s*)?{}\s*", name_match, *SIZE_PROD_MATCH);
+                format!(r"(?:{}\s*:\s*)?{}\s*", AXIS_NAME_MATCH, *SIZE_PROD_MATCH);
             let trailing_comment_match = r"(?:#.*?)?"; // TODO: test this never matches stuff it's not supposed to...
             let full_regex = format!(
                 r"^({ws})(?:(\d+)|{nm})(?: {nm})?(?: \[((?:{sh},\s*)*(?:{sh})?)\])?(?: ([a-zA-Z]+))?(?: (.*?))?(?:{ws}){com}$",
@@ -356,8 +365,8 @@ impl Parser {
                         None
                     }
                 });
-                // TODO: named axes!
-                let shape = flip_op_result(re_captures.get(shape_cap).map(|shape_cap| {
+
+                let (shape, named_axes) = if let Some(shape_cap) = re_captures.get(shape_cap) {
                     let mut axis_strs: Vec<_> =
                         shape_cap.as_str().split(',').map(|z| z.trim()).collect();
                     if axis_strs.last() == Some(&"") {
@@ -366,18 +375,39 @@ impl Parser {
                         axis_strs.pop();
                     }
 
-                    let parse_axis_size = |x| {
-                        SymbolicSizeProduct::parse_to_usize(x).context(concat!(
-                            "failed to parse number (including symbolic product) for shape",
-                            "\nthis is supposed to parse named axes but it currently doesn't!"
+                    let parse_axis_size = |x| -> Result<(usize, Option<&str>)> {
+                        let (striped_x, named_axis) = extract_named_axis(x)?;
+
+                        Ok((
+                            SymbolicSizeProduct::from_str(striped_x)
+                                .context(
+                                    "failed to parse number (including symbolic product) for shape",
+                                )?
+                                .try_into()
+                                .context(
+                                    "failed to convert size product to usize in parse of shape",
+                                )?,
+                            named_axis,
                         ))
                     };
 
-                    axis_strs
+                    let axes_info = axis_strs
                         .into_iter()
                         .map(parse_axis_size)
-                        .collect::<Result<Sv<_>, _>>()
-                }))?;
+                        .collect::<Result<Vec<(usize, Option<&str>)>, _>>()?;
+
+                    let shape = axes_info.iter().map(|(a, _)| *a).collect::<Shape>();
+                    let mut named_axes = NamedAxes::default();
+                    for (i, (_, name)) in axes_info.iter().enumerate() {
+                        if let Some(name_) = name {
+                            named_axes.insert(i.try_into().unwrap(), (*name_).to_owned());
+                        }
+                    }
+
+                    (Some(shape), Some(named_axes))
+                } else {
+                    (None, None)
+                };
                 let variant = flip_op_result(re_captures.get(variant_cap).map(|s| {
                     s.as_str()
                         .parse()
@@ -404,6 +434,7 @@ impl Parser {
                             })?,
                             extra: extra.unwrap_or_else(|| "".to_owned()),
                             children: vec![],
+                            named_axes,
                         },
                     );
                 } else if self.on_repeat_check_info_same {
@@ -613,6 +644,8 @@ impl Parser {
                     &ps.extra[FANCY_PREFIX.len()..],
                     children,
                     ps.name.clone(),
+                ).context(
+                    "Couldn't parse fancy einsum string. If you wanted to parse a regular einsum, remove \"fancy:\".",
                 )
             } else {
                 Einsum::from_einsum_string(&ps.extra, children.clone(), ps.name.clone()).context(
@@ -841,20 +874,50 @@ impl Parser {
                 unimplemented!()
             }
         };
+        if let Some(provided_shape) = &ps.shape {
+            if result.shape() != provided_shape {
+                bail!(ParseCircuitError::PassedInShapeDoesntMatchComputedShape {
+                    provided_shape: provided_shape.clone(),
+                    computed_shape: result.shape().clone(),
+                })
+            }
+        }
+        let result = if let Some(named_axes) = ps.named_axes.clone() {
+            set_named_axes(result, named_axes)?
+        } else {
+            result
+        };
         Ok(result)
     }
 }
 
+fn extract_named_axis<'a>(axis: &'a str) -> Result<(&'a str, Option<&'a str>)> {
+    let trimed = axis.trim();
+    if trimed.contains(':') {
+        let split = trimed.split(":").collect_vec();
+        if split.len() != 2 {
+            bail!(ParseCircuitError::InvalidAxisShapeFormat {
+                axis: axis.to_owned()
+            })
+        }
+        let axis_name = split[0];
+        let striped_axis = split[1];
+        Ok((striped_axis, Some(axis_name)))
+    } else {
+        Ok((axis, None))
+    }
+}
+
 #[apply(python_error_exception)]
-#[base_error_name(ParseArg)]
+#[base_error_name(ReferenceCirc)]
 #[base_exception(PyValueError)]
 #[derive(Error, Debug, Clone)]
-pub enum ParseArgError {
+pub enum ReferenceCircError {
     #[error("circuit={circuit:?} ({e_name})")]
-    ReferenceCircuitByNameHasNoneName { circuit: CircuitRc },
+    ByNameHasNoneName { circuit: CircuitRc },
 
     #[error("dup_idents={dup_idents:?} ({e_name})")]
-    ReferenceCircuitDuplicateIdentifier { dup_idents: HashMap<String, usize> },
+    DuplicateIdentifier { dup_idents: HashMap<String, usize> },
 }
 
 const REGEX_101_MSG: &str = "\nTry using regex101.com : )\n(TODO: auto generate regex test page via https://github.com/firasdib/Regex101/wiki/API#python-3)";
@@ -942,6 +1005,13 @@ pub enum ParseCircuitError {
         line: String,
     },
 
+    #[error("computed_shape={computed_shape:?} != provided_shape={provided_shape:?} ({e_name})")]
+    PassedInShapeDoesntMatchComputedShape {
+        computed_shape: Shape,
+        provided_shape: Shape,
+    },
+
+    // different error type from above for clarity
     #[error(
         "loaded_from_hash_shape={loaded_from_hash_shape:?} != provided_shape={provided_shape:?}.{}\n({e_name})",
         " Note that you don't have to provide the shape if you provide a hash"
@@ -978,4 +1048,7 @@ pub enum ParseCircuitError {
 
     #[error("extra='{extra}' invalid. Should contain | (TODO: example). ({e_name})")]
     StoredCumulantVarExtraInvalid { extra: String },
+
+    #[error("Invalid formating for shape axis {axis} ({e_name})")]
+    InvalidAxisShapeFormat { axis: String },
 }
