@@ -2,12 +2,14 @@ use std::{
     fmt,
     fmt::{Debug, Write},
     hash::{Hash, Hasher},
+    ops::Deref,
     sync::Mutex,
 };
 
 use anyhow::{anyhow, bail, Context, Error, Result};
 use itertools::Itertools;
 use num_bigint::BigUint;
+use once_cell::unsync::Lazy;
 use pyo3::{once_cell::GILLazy, prelude::*};
 use rr_util::{
     fn_struct, py_types::MaybeNotSet, python_println, symbolic_size::SymbolicSizeProduct,
@@ -19,22 +21,91 @@ use crate::{
     circuit_utils::{count_nodes, total_flops},
     generalfunction::SpecTrait,
     opaque_iterative_matcher::{OpaqueIterativeMatcher, OpaqueIterativeMatcherVal},
-    parsing::ParseCircuitError,
+    parsing::{get_reference_circuits, ParseCircuitError},
     prelude::*,
     HashBytes,
 };
 
 // TODO: improve color scheme
-const COLOR_CODES: [usize; 14] = [
-    31, // Red
-    32, // Green
-    33, // Yellow
-    34, // Blue
-    35, // Magenta
-    36, // Cyan
-    90, // White
-    91, 92, 93, 94, 95, 96, 97,
-];
+const COLOR_CODES: [usize; 14] = [31, 32, 33, 34, 35, 36, 90, 91, 92, 93, 94, 95, 96, 97];
+const NAME_TO_COLOR: Lazy<HashMap<String, usize>> = Lazy::new(|| {
+    HashMap::from_iter(
+        [
+            ("Red", 0),
+            ("Green", 1),
+            ("Yellow", 2),
+            ("Blue", 3),
+            ("Magenta", 4),
+            ("Cyan", 5),
+            ("White", 6),
+        ]
+        .into_iter()
+        .map(|(a, b)| (a.to_owned(), b)),
+    )
+});
+
+#[derive(Clone, Copy, Hash, PartialEq, Eq, PartialOrd, Ord)]
+pub struct CliColor(Option<usize>);
+
+impl CliColor {
+    pub const NONE: Self = CliColor(None);
+    pub fn from_string(string: String) -> Result<CliColor> {
+        NAME_TO_COLOR
+            .get(&string)
+            .ok_or_else(|| anyhow!("unknown color name {}", string))
+            .map(|x| CliColor(Some(*x)))
+    }
+    pub fn new(int: usize) -> Self {
+        CliColor(Some(int % COLOR_CODES.len()))
+    }
+}
+impl<'source> FromPyObject<'source> for CliColor {
+    fn extract(circuit_obj: &'source PyAny) -> PyResult<Self> {
+        {
+            if circuit_obj.is_none() {
+                return Ok(CliColor(None));
+            }
+            circuit_obj
+                .clone()
+                .extract::<Option<usize>>()
+                .map(|o| CliColor(o))
+                .or_else(|_e| CliColor::from_string(circuit_obj.clone().extract()?))
+                .map_err(|e| e.into())
+        }
+    }
+}
+
+impl IntoPy<PyObject> for CliColor {
+    fn into_py(self, py: Python<'_>) -> PyObject {
+        {
+            self.0.into_py(py)
+        }
+    }
+}
+
+impl From<Option<CliColor>> for CliColor {
+    fn from(value: Option<CliColor>) -> Self {
+        match value {
+            None => CliColor::NONE,
+            Some(x) => x,
+        }
+    }
+}
+
+impl Deref for CliColor {
+    type Target = Option<usize>;
+
+    fn deref(&self) -> &Self::Target {
+        &self.0
+    }
+}
+
+#[macro_export]
+macro_rules! clicolor {
+    ($name:ident) => {
+        CliColor::from_string(stringify!($name).to_owned()).unwrap()
+    };
+}
 
 const BAR: &'static str = "│";
 const TEE: &'static str = "├";
@@ -43,7 +114,7 @@ const UP_ELBOW: &'static str = "└";
 const DOWN_ELBOW: &'static str = "┌";
 const HORIZ_BAR: &'static str = "─";
 
-fn_struct!(pub CircuitToColorCode:Fn(circuit:CircuitRc)->Option<usize>);
+fn_struct!(pub CircuitToColorCode:Fn(circuit:CircuitRc)->CliColor);
 fn_struct!(pub CircuitCommenter:Fn(circuit:CircuitRc)->String);
 
 #[pyclass]
@@ -51,6 +122,8 @@ fn_struct!(pub CircuitCommenter:Fn(circuit:CircuitRc)->String);
 pub struct PrintOptions {
     #[pyo3(get, set)]
     pub bijection: bool,
+    #[pyo3(get)]
+    pub reference_circuits: HashMap<CircuitRc, String>, /* TODO: make a checking setter for me as needed */
     #[pyo3(get, set)]
     pub shape_only_when_necessary: bool,
     pub traversal: Option<OpaqueIterativeMatcherVal>,
@@ -74,6 +147,7 @@ pub struct PrintOptions {
     pub seen_children_same_line: bool,
     #[pyo3(get, set)]
     pub only_child_below: bool,
+    #[pyo3(get, set)]
     pub tensor_index_literal: bool,
 }
 
@@ -81,6 +155,7 @@ impl Default for PrintOptions {
     fn default() -> Self {
         Self {
             bijection: true,
+            reference_circuits: Default::default(),
             shape_only_when_necessary: true,
             traversal: None,
             leaves_on_top: false,
@@ -185,6 +260,8 @@ impl PrintOptions {
     #[new]
     #[args(
         bijection = "PrintOptions::default().bijection",
+        reference_circuits = "PrintOptions::default().reference_circuits",
+        reference_circuits_by_name = "vec![]",
         shape_only_when_necessary = "PrintOptions::default().shape_only_when_necessary",
         traversal = "None",
         leaves_on_top = "PrintOptions::default().leaves_on_top",
@@ -201,6 +278,8 @@ impl PrintOptions {
     )]
     pub fn new(
         bijection: bool,
+        reference_circuits: HashMap<CircuitRc, String>,
+        reference_circuits_by_name: Vec<CircuitRc>,
         shape_only_when_necessary: bool,
         traversal: Option<OpaqueIterativeMatcherVal>,
         leaves_on_top: bool,
@@ -217,6 +296,16 @@ impl PrintOptions {
     ) -> Result<Self> {
         let result = Self {
             bijection,
+            reference_circuits: get_reference_circuits(
+                reference_circuits
+                    .into_iter()
+                    .map(|(a, b)| (b, a))
+                    .collect(),
+                reference_circuits_by_name,
+            )?
+            .into_iter()
+            .map(|(a, b)| (b, a))
+            .collect(),
             shape_only_when_necessary,
             traversal,
             leaves_on_top,
@@ -237,6 +326,7 @@ impl PrintOptions {
 
     #[args(
         bijection = "Default::default()",
+        reference_circuits = "Default::default()",
         shape_only_when_necessary = "Default::default()",
         traversal = "Default::default()",
         leaves_on_top = "Default::default()",
@@ -254,6 +344,7 @@ impl PrintOptions {
     pub fn evolve(
         &self,
         bijection: MaybeNotSet<bool>,
+        reference_circuits: MaybeNotSet<HashMap<CircuitRc, String>>,
         shape_only_when_necessary: MaybeNotSet<bool>,
         traversal: MaybeNotSet<Option<OpaqueIterativeMatcherVal>>,
         leaves_on_top: MaybeNotSet<bool>,
@@ -271,6 +362,8 @@ impl PrintOptions {
         let cloned = self.clone();
         Self::new(
             bijection.0.unwrap_or(cloned.bijection),
+            reference_circuits.0.unwrap_or(cloned.reference_circuits),
+            vec![],
             shape_only_when_necessary
                 .0
                 .unwrap_or(cloned.shape_only_when_necessary),
@@ -345,6 +438,9 @@ impl PrintOptions {
                 "bijection doesn't support seen children same line, but technically could"
             ))
         }
+        if self.bijection && self.tensor_index_literal {
+            bail!(anyhow!("bijection doesn't support index tensor literal"))
+        }
         if self.seen_children_same_line && self.comment_arg_names {
             bail!(anyhow!(
                 "children same line + comment arg names not supported"
@@ -387,7 +483,7 @@ impl PrintOptions {
                 (0..circ.info().rank())
                     .map(|x| match circ.info().named_axes.get(&(x as u8)) {
                         None => SymbolicSizeProduct::from(circ.shape()[x]).to_string(),
-                        Some(s) => format!("{}:{}", self.string_escape(s), circ.shape()[x]),
+                        Some(s) => format!("{}:{}", s, circ.shape()[x]),
                     })
                     .collect::<Vec<_>>()
                     .join(",")
@@ -403,11 +499,22 @@ impl PrintOptions {
                 Circuit::Scalar(scalar) => {
                     format!("{:.}", scalar.value)
                 }
-                Circuit::Rearrange(rearrange) => rearrange.spec.to_string(),
-                Circuit::Einsum(einsum) => einsum.get_spec().to_einsum_string(),
+                Circuit::Rearrange(rearrange) => rearrange.spec.to_maybe_fancy_einops_string(
+                    rearrange.node.info().named_axes.clone(),
+                    rearrange.info().named_axes.clone(),
+                ),
+                Circuit::Einsum(einsum) => einsum.get_spec().to_maybe_fancy_string(
+                    einsum.info().named_axes.clone(),
+                    einsum
+                        .children()
+                        .map(|c| c.info().named_axes.clone())
+                        .collect(),
+                ),
                 Circuit::Index(index) => {
                     if self.bijection {
                         index.index.repr_bijection(self.tensor_index_literal)?
+                    } else if self.tensor_index_literal {
+                        index.index.repr_bijection(self.tensor_index_literal)?
                     } else {
                         format!("{}", index.index)
                     }
@@ -438,7 +545,7 @@ impl PrintOptions {
                     }
                 }
                 Circuit::Module(mn) => {
-                    if mn.spec.arg_specs.len() >= 3
+                    if mn.spec.arg_specs.len() >= 2
                         && mn
                             .spec
                             .arg_specs
@@ -494,7 +601,7 @@ impl PrintOptions {
             if circ.info().numel() > BigUint::from(threshold) && !circ.is_array() {
                 result.push_str(&format!(
                     " # {}",
-                    color(&oom_fmt(circ.info().numel()), Some(0))
+                    color(&oom_fmt(circ.info().numel()), clicolor!(Red))
                 ));
             }
         }
@@ -573,10 +680,15 @@ impl PrintOptions {
                 result.push('\n');
                 return Ok(());
             }
+            if let Some(ident) = selfy.reference_circuits.get(&circ.clone().rc()) {
+                result.push_str(&selfy.string_escape(ident));
+                result.push('\n');
+                return Ok(());
+            }
 
             let mut used_color = false;
             if let Some(colorer) = &selfy.colorer {
-                if let Some(color) = colorer.call(circ.clone().rc())? {
+                if let Some(color) = colorer.call(circ.clone().rc())?.0 {
                     used_color = true;
                     result.push_str(&format!(
                         "\u{001b}[{}m",
@@ -799,18 +911,18 @@ impl PrintOptions {
         CircuitToColorCode::new_dyn(Box::new(|circuit: CircuitRc| {
             let mut hasher = rustc_hash::FxHasher::default();
             circuit.variant_string().hash(&mut hasher);
-            Ok(Some(hasher.finish() as usize))
+            Ok(CliColor::new(hasher.finish() as usize))
         }))
     }
 
     #[staticmethod]
     pub fn hash_colorer() -> CircuitToColorCode {
         CircuitToColorCode::new_dyn(Box::new(|circuit: CircuitRc| {
-            Ok(Some(circuit.info().hash_usize()))
+            Ok(CliColor::new(circuit.info().hash_usize()))
         }))
     }
     #[staticmethod]
-    pub fn fixed_color(color: Option<usize>) -> CircuitToColorCode {
+    pub fn fixed_color(color: CliColor) -> CircuitToColorCode {
         CircuitToColorCode::new_dyn(Box::new(move |_circuit: CircuitRc| Ok(color)))
     }
 
@@ -818,6 +930,27 @@ impl PrintOptions {
     pub fn computability_colorer() -> CircuitToColorCode {
         CircuitToColorCode::new_dyn(Box::new(|c| Ok(circuit_to_computability_color_code(c))))
     }
+
+    #[staticmethod]
+    pub fn circuit_set_colorer(
+        map: HashMap<CliColor, HashSet<CircuitRc>>,
+    ) -> Result<CircuitToColorCode> {
+        let map: HashMap<HashBytes, CliColor> = map
+            .into_iter()
+            .map(|(color, c)| {
+                Ok(c.into_iter()
+                    .map(|c| (c.info().hash, color))
+                    .collect::<Vec<_>>())
+            })
+            .collect::<Result<Vec<_>>>()?
+            .into_iter()
+            .flatten()
+            .collect();
+
+        Ok(CircuitToColorCode::new_dyn(Box::new(move |c| {
+            Ok(map.get(&c.info().hash).cloned().into())
+        })))
+    }
 }
 
 fn init_print_options() -> PrintOptions {
@@ -837,28 +970,28 @@ fn init_print_options() -> PrintOptions {
     PrintOptions::debug_default(end_depth)
 }
 
-pub fn circuit_to_size_color_code(circuit: CircuitRc) -> Option<usize> {
+pub fn circuit_to_size_color_code(circuit: CircuitRc) -> CliColor {
     let size = circuit.info().numel_usize();
     if size == usize::MAX {
-        Some(2)
+        clicolor!(Magenta)
     } else if size > 10_000_000_000 {
-        Some(3)
+        clicolor!(Red)
     } else if size > 300_000_000 {
-        Some(0)
+        clicolor!(Yellow)
     } else if size > 300_000 {
-        Some(1)
+        clicolor!(Green)
     } else {
-        None
+        CliColor::NONE
     }
 }
 
-pub fn circuit_to_computability_color_code(circuit: CircuitRc) -> Option<usize> {
+pub fn circuit_to_computability_color_code(circuit: CircuitRc) -> CliColor {
     if !circuit.info().can_be_sampled {
-        Some(1)
+        clicolor!(Green)
     } else if !circuit.info().is_explicitly_computable {
-        Some(2)
+        clicolor!(Yellow)
     } else {
-        None
+        CliColor::NONE
     }
 }
 
@@ -907,8 +1040,8 @@ pub fn print_circuit_stats(circuit: &Circuit) {
     println!("{}", result);
 }
 
-pub fn color(string: &str, color_int: Option<usize>) -> String {
-    if let Some(color_int) = color_int {
+pub fn color(string: &str, color_int: CliColor) -> String {
+    if let Some(color_int) = color_int.0 {
         return format!(
             "\u{001b}[{}m{}\u{001b}[0m",
             COLOR_CODES[color_int % COLOR_CODES.len()],
