@@ -1129,7 +1129,7 @@ pub fn einsum_concat_to_add(einsum: &Einsum) -> Option<Add> {
 }
 
 #[pyfunction]
-pub fn index_split_axes(node: &Index, top_axes: std::collections::HashSet<usize>) -> Option<Index> {
+pub fn index_split_axes(node: &Index, top_axes: HashSet<usize>) -> Option<Index> {
     let mut bottom: Vec<TensorAxisIndex> = vec![];
     let mut top: Vec<TensorAxisIndex> = vec![];
     if top_axes.iter().any(|i| *i >= node.node.info().rank()) {
