@@ -1,7 +1,4 @@
-use std::{
-    iter::zip,
-    simd::{u64x8, u8x8, SimdPartialEq, SimdUint},
-};
+use std::simd::{u64x8, u8x8, SimdPartialEq};
 
 use anyhow::{bail, Result};
 use circuit_base::{
@@ -16,7 +13,7 @@ use rr_util::{
     python_error_exception, timed, timed_value,
     util::{apply_fn_until_none, apply_fn_until_same, mapping_until_end, AsOp, HashBytes},
 };
-use rustc_hash::{FxHashMap as HashMap, FxHashSet as HashSet};
+use rustc_hash::FxHashMap as HashMap;
 use thiserror::Error;
 
 use crate::{
