@@ -4,7 +4,7 @@ use anyhow::{bail, Context, Result};
 use circuit_base::{
     circuit_utils::{hash_to_node_non_free, toposort_circuit},
     get_compatible_dtype,
-    module::inline_all_modules,
+    module::substitute_all_modules,
     prelude::*,
     print::{oom_fmt, PrintOptions},
     visit_circuit_non_free, Add, Array, IrreducibleNode, Leaf, ModuleSpec, Scalar, Symbol,
@@ -753,7 +753,7 @@ pub fn circuit_to_schedule(
 }
 
 pub fn circuit_to_schedule_naive_toposort(circuit: CircuitRc) -> Schedule {
-    let circuit = inline_all_modules(circuit);
+    let circuit = substitute_all_modules(circuit);
     let toposorted = toposort_circuit(circuit.clone());
     let (mut result, kept_keys) = order_to_schedule(
         &toposorted,
