@@ -20,7 +20,7 @@ use rr_util::{
     impl_both_by_big_hash, python_error_exception, setup_callable, simple_default, simple_from,
     tensor_util::Slice,
     util::{
-        arc_ref_clone, arc_unwrap_or_clone, as_sorted, flip_op_result, split_first_take,
+        arc_ref_clone, arc_unwrap_or_clone, as_sorted, flip_op_result, split_first_take, transpose,
         EmptySingleMany as ESM, HashBytes,
     },
 };
@@ -51,6 +51,7 @@ pub enum IterativeMatcherData {
     SpecCircuit(IterativeMatcherRc),
     NoModuleSpec(IterativeMatcherRc),
     Chains(BTreeSet<ChainItem>),
+    All(Vec<IterativeMatcherRc>),
     Raw(RawIterativeMatcher),
     PyFunc(PyObject),
 }
@@ -156,6 +157,7 @@ impl IterativeMatcherData {
             Self::SpecCircuit(_) => uuid!("b728a8c3-0493-4ee5-96c2-71e4900fc0f8"),
             Self::NoModuleSpec(_) => uuid!("b728a8c3-0493-4ee5-96c2-71e4900fc0f8"),
             Self::Chains(_) => uuid!("958d03ed-7a1a-4ea9-8dd4-d4a8a68feecb"),
+            Self::All(_) => uuid!("b4079909-ad3e-4a32-a016-43c96f262237"),
             Self::Raw(_) => uuid!("5838ac96-0f48-4cdb-874f-d4f68ce3a52b"),
             Self::PyFunc(_) => uuid!("d3afc3c0-5c86-46df-9e41-7944caedd901"),
         }
@@ -201,6 +203,11 @@ impl IterativeMatcherData {
                     hasher.update(&chain.hash);
                 }
             }
+            Self::All(matchers) => {
+                for matcher in matchers {
+                    hasher.update(&matcher.hash);
+                }
+            }
             Self::Raw(x) => {
                 hasher.update(&(Arc::as_ptr(&x.0) as *const () as usize).to_le_bytes());
             }
@@ -452,6 +459,7 @@ pub fn require_single(x: UpdatedIterativeMatcher) -> Result<Update> {
     }
 }
 
+/// returned option inside of vec is whether or not finished
 pub fn per_child(
     updated: Option<UpdatedIterativeMatcher>,
     matcher: IterativeMatcherRc,
@@ -564,6 +572,10 @@ impl IterativeMatcher {
         Self::any(vec![self.crc(), other])
     }
 
+    pub fn and(self, other: IterativeMatcherRc) -> Self {
+        Self::all(vec![self.crc(), other])
+    }
+
     // TODO: this is kinda shitty, see https://github.com/redwoodresearch/unity/issues/1524
     pub fn validate_matched(&self, matched: &BTreeSet<CircuitRc>) -> Result<()> {
         match &self.data {
@@ -579,6 +591,7 @@ impl IterativeMatcher {
                 }
                 Ok(())
             }
+            IterativeMatcherData::All(_) => Ok(()), // TODO
             IterativeMatcherData::PyFunc(_) | IterativeMatcherData::Raw(_) => Ok(()),
         }
     }
@@ -845,6 +858,60 @@ impl IterativeMatcher {
                     found: any_chain_finished,
                 }
             }
+            IterativeMatcherData::All(matchers) => {
+                let results = matchers
+                    .iter()
+                    .map(|x| x.match_iterate(circuit.clone()))
+                    .collect::<Result<Vec<_>>>()?;
+                let found = results.iter().all(|x| x.found);
+                let all_updated: Vec<_> = results.into_iter().map(|x| x.updated).collect();
+
+                if all_updated.iter().all(|x| x.is_none()) {
+                    IterateMatchResults {
+                        found,
+                        updated: None,
+                    }
+                } else if all_updated.iter().any(|x| all_finished(&x)) {
+                    IterateMatchResults {
+                        found,
+                        updated: Some(None.into()),
+                    }
+                } else if all_updated.iter().all(|x| {
+                    x.as_ref()
+                        .map(|x| matches!(x, UpdatedIterativeMatcher::Single(_)))
+                        .unwrap_or(true)
+                }) {
+                    let updated_sub: Vec<_> = all_updated
+                        .into_iter()
+                        .zip(matchers)
+                        .map(|(x, matcher)| {
+                            x.map(|x| match x {
+                                UpdatedIterativeMatcher::Single(x) => x.0.unwrap(),
+                                UpdatedIterativeMatcher::Many(_) => unreachable!(),
+                            })
+                            .unwrap_or(matcher.clone())
+                        })
+                        .collect();
+                    let updated = Some(Some(Self::all(updated_sub).rc()).into());
+
+                    IterateMatchResults { found, updated }
+                } else {
+                    let all_per_child = all_updated
+                        .into_iter()
+                        .zip(matchers)
+                        .map(|(updated, matcher)| per_child(updated, matcher.clone(), num_children))
+                        .collect();
+                    let updated = transpose(all_per_child, num_children)
+                        .into_iter()
+                        .map(|new_matchers| {
+                            Some(Self::all(new_matchers.into_iter().collect::<Option<_>>()?).rc())
+                        })
+                        .collect::<Vec<_>>();
+                    let updated = Some(updated.into());
+
+                    IterateMatchResults { found, updated }
+                }
+            }
             IterativeMatcherData::Raw(f) => f.0(circuit)?,
             IterativeMatcherData::PyFunc(pyfunc) => {
                 Python::with_gil(|py| pyfunc.call1(py, (circuit,)).and_then(|r| r.extract(py)))?
@@ -883,6 +950,12 @@ impl IterativeMatcher {
         .into()
     }
 
+    #[staticmethod]
+    #[args(matchers = "*")]
+    pub fn all(matchers: Vec<IterativeMatcherRc>) -> Self {
+        IterativeMatcherData::All(matchers).into()
+    }
+
     #[staticmethod]
     #[args(first, rest = "*")]
     pub fn new_chain(first: IterativeMatcherRc, rest: Vec<IterativeMatcherRc>) -> Self {
@@ -935,6 +1008,13 @@ impl IterativeMatcher {
         arc_unwrap_or_clone(other.0).or(self.crc())
     }
 
+    fn __and__(&self, other: IterativeMatcherRc) -> Self {
+        self.clone().and(other)
+    }
+    fn __rand__(&self, other: IterativeMatcherRc) -> Self {
+        arc_unwrap_or_clone(other.0).and(self.crc())
+    }
+
     // TODO: write flatten/simplify method if we want the extra speed + niceness!
 }
 
