@@ -6,13 +6,13 @@ use std::{
 
 use anyhow::Result;
 use circuit_base::{
-    print::{color, last_child_arrows, PrintOptions},
+    clicolor,
+    print::{color, last_child_arrows, CliColor, PrintOptions},
     CircuitNode, CircuitNodeUnion, CircuitRc,
 };
 use itertools::Itertools;
-use macro_rules_attribute::apply;
 use pyo3::prelude::*;
-use rr_util::{cached_lambda, python_println};
+use rr_util::python_println;
 use rustc_hash::FxHashSet as HashSet;
 
 use crate::{
@@ -189,13 +189,13 @@ pub fn repr_matcher_debug(
             last_child_arrows(&last_child_stack, true, print_options.arrows),
             print_options.repr_line_info(circuit.clone())?,
             if matched.found {
-                color(&" # Found", Some(1))
+                color(&" # Found", clicolor!(Green))
             } else {
                 "".to_owned()
             },
             color(
                 &format!(" # {}", &matcher.0),
-                Some(matcher.__hash__() as usize)
+                CliColor::new(matcher.__hash__() as usize)
             ),
         );
         result.push_str(&line_repr);
