@@ -1,4 +1,4 @@
-use std::collections::BTreeSet;
+use std::collections::{BTreeMap, BTreeSet};
 
 use anyhow::{anyhow, bail, Result};
 use circuit_base::{expand_node::expand_node, CircuitNode, CircuitRc};
@@ -462,7 +462,14 @@ pub struct Expander {
     pub default_fancy_validate: bool,
     #[pyo3(get)]
     pub(super) suffix: Option<String>,
-    pub(super) batch_cache: cached::UnboundCache<(HashBytes, Vec<IterativeMatcherRc>), CircuitRc>,
+    pub(super) batch_cache: cached::UnboundCache<
+        (
+            HashBytes,
+            Vec<IterativeMatcherRc>,
+            BTreeMap<CircuitRc, CircuitRc>,
+        ),
+        CircuitRc,
+    >,
     pub(super) validation_getter: Getter,
 }
 
@@ -525,20 +532,21 @@ impl Expander {
                     .validate(circuit.clone(), m.clone())?;
             }
         }
-        self.batch_impl(circuit, self.matchers.clone())
+        self.batch_impl(circuit, self.matchers.clone(), &Default::default())
     }
 }
 
 impl Expander {
     #[apply(cached_method)]
     #[self_id(self_)]
-    #[key((circuit.info().hash, matchers.clone()))]
+    #[key((circuit.info().hash, matchers.clone(), extra_replacements.clone()))]
     #[use_try]
     #[cache_expr(batch_cache)]
     fn batch_impl(
         &mut self,
         circuit: CircuitRc,
         matchers: Vec<IterativeMatcherRc>,
+        extra_replacements: &BTreeMap<CircuitRc, CircuitRc>,
     ) -> Result<CircuitRc> {
         let results = matchers
             .iter()
@@ -557,6 +565,9 @@ impl Expander {
 
             return self_.replacements[idx].run(circuit);
         }
+        if let Some(replaced) = extra_replacements.get(&circuit) {
+            return Ok(replaced.clone());
+        }
 
         if results.iter().all(|x| all_finished(&x.updated)) {
             return Ok(circuit);
@@ -574,11 +585,22 @@ impl Expander {
 
         let new_children = circuit
             .children()
-            .zip(new_matchers_per)
-            .map(|(c, new_matchers)| self_.batch_impl(c, new_matchers))
+            .zip(new_matchers_per.clone())
+            .map(|(c, new_matchers)| self_.batch_impl(c, new_matchers, extra_replacements))
             .collect::<Result<_>>()?;
 
-        Ok(expand_node(circuit, &new_children)?.add_suffix(self_.suffix.as_ref().map(|x| &**x)))
+        Ok(expand_node(circuit, &new_children, |c, rep, child_idx| {
+            self_.batch_impl(
+                c,
+                new_matchers_per[child_idx].clone(),
+                &extra_replacements
+                    .iter()
+                    .map(|(a, b)| (a.clone(), b.clone()))
+                    .chain(rep)
+                    .collect(),
+            )
+        })?
+        .add_suffix(self_.suffix.as_ref().map(|x| &**x)))
     }
 
     pub fn suffix(&self) -> Option<&str> {
