@@ -151,13 +151,13 @@ einsum = lambda tensors_and_axes, out_axes: torch.einsum(*itertools.chain(*tenso
 def random_indices(input: torch.Tensor, shape, replacement: bool, seed: Optional[int] = None) -> torch.Tensor:
     shape = tuple(shape)
 
-    if seed is not None:
-        torch.manual_seed(seed)
+    generator: Optional[torch.Generator] = torch.Generator().manual_seed(seed) if seed is not None else None
 
     return torch.multinomial(
         input,
         num_samples=math.prod(shape),
         replacement=replacement,
+        generator=generator,
     ).reshape(shape)
 
 
