@@ -24,7 +24,7 @@ use crate::{
     tensor_util::{check_canon_idxs, PyParseError, Shape},
     util::{
         dict_to_list, filter_out_idx, is_unique, vec_map_insert, AxisInt, EinsumAxes, HashBytes,
-        ALPHABET,
+        NamedAxes, ALPHABET,
     },
 };
 /// OpSize is a memory optimization of Option<usize> that stores a 63-bit int and one "is this none" bit
@@ -554,6 +554,63 @@ impl RearrangeSpec {
         )
     }
 
+    /// Only supports a small subsets of rearrange specs. Return None if not supported.
+    /// Self's named_axes preceeds over node's named axes.
+    /// Does not check for conflicts between different names.
+    pub fn to_fancy_einops_string(
+        &self,
+        node_named_axes: NamedAxes,
+        self_named_axes: NamedAxes,
+    ) -> Option<String> {
+        let input_int_unpacked = self
+            .input_ints
+            .iter()
+            .map(|v| if v.len() == 1 { Some(v[0]) } else { None })
+            .collect::<Option<Vec<AxisInt>>>()?;
+        let output_int_unpacked = self
+            .output_ints
+            .iter()
+            .map(|v| if v.len() == 1 { Some(v[0]) } else { None })
+            .collect::<Option<Vec<AxisInt>>>()?;
+        let mut axis_to_name: HashMap<AxisInt, &String> = Default::default();
+        for (i, a) in input_int_unpacked.iter().enumerate() {
+            if let Some(name) = node_named_axes.get(&(i as u8)) {
+                axis_to_name.insert(*a, name);
+            }
+        }
+        for (i, a) in output_int_unpacked.iter().enumerate() {
+            if let Some(name) = self_named_axes.get(&(i as u8)) {
+                axis_to_name.insert(*a, name);
+            }
+        }
+        if axis_to_name.len() > HashSet::from_iter(axis_to_name.values().cloned()).len() {
+            return None; // Can't tolerate duplicate names
+        }
+        let inp_words = input_int_unpacked
+            .iter()
+            .map(|a| Some((*axis_to_name.get(a)?).clone()))
+            .collect::<Option<Vec<String>>>()?
+            .join(" ");
+        let out_words = output_int_unpacked
+            .iter()
+            .map(|a| Some((*axis_to_name.get(a)?).clone()))
+            .collect::<Option<Vec<String>>>()?
+            .join(" ");
+        Some(inp_words + " -> " + &out_words)
+    }
+
+    pub fn to_maybe_fancy_einops_string(
+        &self,
+        node_named_axes: NamedAxes,
+        self_named_axes: NamedAxes,
+    ) -> String {
+        if let Some(fancy_string) = self.to_fancy_einops_string(node_named_axes, self_named_axes) {
+            fancy_string
+        } else {
+            self.to_string()
+        }
+    }
+
     #[staticmethod]
     #[pyo3(name = "from_string")]
     pub fn py_from_string(string: &str) -> Result<Self> {
