@@ -1,4 +1,5 @@
 use std::{
+    borrow::Borrow,
     cmp::{max, min},
     fmt::{self, Display},
     iter::zip,
@@ -20,7 +21,7 @@ use smallvec::SmallVec as Sv;
 use thiserror::Error;
 use uuid::uuid;
 
-use super::py_types::ExtraPySelfOps;
+use super::py_types::{ExtraPySelfOps, PyShape};
 use crate::{
     atr, filter_by_variant,
     lru_cache::TensorCacheRrfs,
@@ -31,11 +32,17 @@ use crate::{
 };
 pub type Shape = Sv<[usize; 6]>; // TODO: maybe this should be struct which preserves invariant that dim shapes >= 0?
 
-pub fn broadcast_shapes(shapes: &Vec<Shape>) -> Result<Shape> {
+#[crate::pyo3::pyfunction(shapes = "*")]
+#[pyo3(name = "broadcast_shapes")]
+pub fn broadcast_shapes_py(shapes: Vec<Shape>) -> Result<PyShape> {
+    broadcast_shapes(&shapes).map(PyShape)
+}
+
+pub fn broadcast_shapes(shapes: &[impl Borrow<[usize]>]) -> Result<Shape> {
     if shapes.is_empty() {
         return Ok(sv![]);
     }
-    let ranks: Vec<usize> = shapes.iter().map(|x| x.len()).collect();
+    let ranks: Vec<usize> = shapes.iter().map(|x| x.borrow().len()).collect();
     let out_rank = *ranks.iter().max().unwrap();
     let mut result: Shape = sv![1; out_rank];
     for axis_from_end in 1..out_rank + 1 {
@@ -44,14 +51,17 @@ pub fn broadcast_shapes(shapes: &Vec<Shape>) -> Result<Shape> {
             match axis_for_shape {
                 None => {}
                 Some(j) => {
-                    let axis_len = shape[j];
+                    let axis_len = shape.borrow()[j];
                     if axis_len == 0
                         || axis_len != 1
                             && result[out_rank - axis_from_end] != 1
                             && result[out_rank - axis_from_end] != axis_len
                     {
                         bail!(MiscInputError::NotBroadcastable {
-                            shapes: shapes.to_owned(),
+                            shapes: shapes
+                                .iter()
+                                .map(|x| x.borrow().iter().cloned().collect())
+                                .collect(),
                         });
                     }
                     result[out_rank - axis_from_end] =
@@ -703,7 +713,6 @@ impl TensorIndex {
                 .map(str::trim)
                 .filter(|x| !x.is_empty())
                 .map(|x| -> Result<_> {
-                    // TODO: trim whitespace?
                     if let Ok(int) = parse_numeric(x) {
                         return Ok(TensorAxisIndex::Single(int));
                     }
