@@ -1,4 +1,5 @@
 use std::{
+    collections::BTreeMap,
     fmt::Debug,
     hash::{BuildHasher, Hash},
     mem,
@@ -20,6 +21,7 @@ pub type HashBytes = [u8; 32];
 // we could use u64 instead of u8 if we wanted.
 // this is u8, so we want to use multiple of 8 inline so we don't pad
 pub type EinsumAxes = SmallVec<[AxisInt; 8]>;
+pub type NamedAxes = BTreeMap<AxisInt, String>;
 
 #[derive(FromPyObject, Copy, Clone, Debug, Eq, PartialEq, Ord, PartialOrd)]
 pub struct HashBytesToPy(pub HashBytes);
