@@ -1,7 +1,8 @@
 use circuit_base::{
     expand_node::ExpandError,
     generalfunction::GeneralFunctionShapeError,
-    parsing::{ParseArgError, ParseCircuitError},
+    named_axes::NamedAxesError,
+    parsing::{ParseCircuitError, ReferenceCircError},
     ConstructError, TensorEvalError,
 };
 use circuit_rewrites::{
@@ -66,9 +67,10 @@ setup_errors!(
     DistributeError,
     SimpConfigError,
     GeneralFunctionShapeError,
-    ParseArgError,
+    ReferenceCircError,
     SymbolicSizeOverflowError,
     SymbolicSizeSetError,
     IterativeMatcherError,
-    ModuleBindError
+    ModuleBindError,
+    NamedAxesError,
 );
